@@ -9,6 +9,11 @@
  * Daniel.Veillard@imag.fr
  */
 
+/*
+ * TODO: handle pathological cases like *[*[@a="b"]]
+ * TODO: detect [number] at compilation, optimize accordingly
+ */
+
 #include "xsltconfig.h"
 
 #include <string.h>
@@ -27,7 +32,7 @@
 #include "keys.h"
 #include "pattern.h"
 
-/* #define DEBUG_PARSING */
+#define DEBUG_PATTERN
 
 /*
  * Types are private:
@@ -60,6 +65,12 @@ struct _xsltStepOp {
     xmlChar *value;
     xmlChar *value2;
     xmlChar *value3;
+    /*
+     * Optimisations for count
+     */
+    xmlNodePtr previous;
+    int        index;
+    int        len;
 };
 
 struct _xsltCompMatch {
@@ -80,6 +91,8 @@ typedef xsltParserContext *xsltParserContextPtr;
 struct _xsltParserContext {
     const xmlChar *cur;			/* the current char being parsed */
     const xmlChar *base;		/* the full expression */
+    xmlDocPtr      doc;			/* the source document */
+    xmlNodePtr    elem;			/* the source element */
     int error;				/* error code */
     xsltCompMatchPtr comp;		/* the result */
 };
@@ -333,10 +346,10 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
             case XSLT_OP_END:
 		return(1);
             case XSLT_OP_ROOT:
-		if ((node->type != XML_DOCUMENT_NODE) &&
-		    (node->type != XML_HTML_DOCUMENT_NODE))
-		    return(0);
-		continue;
+		if ((node->type == XML_DOCUMENT_NODE) ||
+		    (node->type == XML_HTML_DOCUMENT_NODE))
+		    continue;
+		return(0);
             case XSLT_OP_ELEM:
 		if (node->type != XML_ELEMENT_NODE)
 		    return(0);
@@ -493,24 +506,135 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
 		    (node->type == XML_ELEMENT_NODE) &&
 		    (node->parent != NULL)) {
 
-		    /* TODO: cache those informations ?!? */
-		    xmlNodePtr siblings = node->parent->children;
+		    if ((select->previous != NULL) &&
+			(select->previous->parent == node->parent)) {
+			/*
+			 * just walk back to adjust the index
+			 */
+			int i = 0;
+			xmlNodePtr sibling = node;
 
-		    while (siblings != NULL) {
-			if (siblings->type == XML_ELEMENT_NODE) {
-			    if (siblings == node) {
-				len++;
-				pos = len;
-			    } else if (xmlStrEqual(node->name,
-				       siblings->name)) {
-				len++;
+			while (sibling != NULL) {
+			    if (sibling == select->previous)
+				break;
+			    if (xmlStrEqual(node->name, sibling->name)) {
+				if ((select->value2 == NULL) ||
+				    ((sibling->ns != NULL) &&
+				     (xmlStrEqual(select->value2,
+						  sibling->ns->href))))
+				    i++;
+			    }
+			    sibling = sibling->prev;
+			}
+			if (sibling == NULL) {
+			    /* hum going backward in document order ... */
+			    i = 0;
+			    sibling = node;
+			    while (sibling != NULL) {
+				if (sibling == select->previous)
+				    break;
+				if ((select->value2 == NULL) ||
+				    ((sibling->ns != NULL) &&
+				     (xmlStrEqual(select->value2,
+						  sibling->ns->href))))
+				    i--;
+				sibling = sibling->next;
 			    }
 			}
-			siblings = siblings->next;
+			if (sibling != NULL) {
+			    pos = select->index + i;
+			    len = select->len;
+			    select->previous = node;
+			    select->index = pos;
+			} else
+			    pos = 0;
+		    } else {
+			/*
+			 * recompute the index
+			 */
+			xmlNodePtr siblings = node->parent->children;
+
+			while (siblings != NULL) {
+			    if (siblings->type == XML_ELEMENT_NODE) {
+				if (siblings == node) {
+				    len++;
+				    pos = len;
+				} else if (xmlStrEqual(node->name,
+					   siblings->name)) {
+				    if ((select->value2 == NULL) ||
+					((siblings->ns != NULL) &&
+					 (xmlStrEqual(select->value2,
+						      siblings->ns->href))))
+					len++;
+				}
+			    }
+			    siblings = siblings->next;
+			}
 		    }
 		    if (pos != 0) {
 			ctxt->xpathCtxt->contextSize = len;
 			ctxt->xpathCtxt->proximityPosition = pos;
+			select->previous = node;
+			select->index = pos;
+			select->len = len;
+		    }
+		} else if ((select != NULL) && (select->op == XSLT_OP_ALL)) {
+		    if ((select->previous != NULL) &&
+			(select->previous->parent == node->parent)) {
+			/*
+			 * just walk back to adjust the index
+			 */
+			int i = 0;
+			xmlNodePtr sibling = node;
+
+			while (sibling != NULL) {
+			    if (sibling == select->previous)
+				break;
+			    if (sibling->type == XML_ELEMENT_NODE)
+				i++;
+			    sibling = sibling->prev;
+			}
+			if (sibling == NULL) {
+			    /* hum going backward in document order ... */
+			    i = 0;
+			    sibling = node;
+			    while (sibling != NULL) {
+				if (sibling == select->previous)
+				    break;
+				if (sibling->type == XML_ELEMENT_NODE)
+				    i--;
+				sibling = sibling->next;
+			    }
+			}
+			if (sibling != NULL) {
+			    pos = select->index + i;
+			    len = select->len;
+			    select->previous = node;
+			    select->index = pos;
+			} else
+			    pos = 0;
+		    } else {
+			/*
+			 * recompute the index
+			 */
+			xmlNodePtr siblings = node->parent->children;
+
+			while (siblings != NULL) {
+			    if (siblings->type == XML_ELEMENT_NODE) {
+				len++;
+				if (siblings == node) {
+				    pos = len;
+				}
+			    }
+			    siblings = siblings->next;
+			}
+		    }
+		    if (pos != 0) {
+			ctxt->xpathCtxt->contextSize = len;
+			ctxt->xpathCtxt->proximityPosition = pos;
+			select->previous = node;
+			select->index = pos;
+			select->len = len;
 		    }
 		}
 		oldNode = ctxt->node;
@@ -889,6 +1013,9 @@ error:
 void
 xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
     xmlChar *name = NULL;
+    xmlChar *prefix = NULL;
+    xmlChar *ncname = NULL;
+    xmlChar *URL = NULL;
 
     SKIP_BLANKS;
     if ((token == NULL) && (CUR == '@')) {
@@ -922,6 +1049,8 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 	    goto error;
 	}
     }
+
+
     SKIP_BLANKS;
     if (CUR == '(') {
 	xsltCompileIdKeyPattern(ctxt, token, 0);
@@ -937,7 +1066,6 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 	}
 	NEXT;
 	if (xmlStrEqual(token, (const xmlChar *) "child")) {
-	    /* TODO: handle namespace */
 	    name = xsltScanName(ctxt);
 	    if (name == NULL) {
 		xsltGenericError(xsltGenericErrorContext,
@@ -945,9 +1073,26 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 		ctxt->error = 1;
 		goto error;
 	    }
-	    PUSH(XSLT_OP_CHILD, name, NULL);
+	    ncname = xmlSplitQName2(name, &prefix);
+	    if (ncname != NULL) {
+		if (prefix != NULL) {
+		    xmlNsPtr ns;
+
+		    ns = xmlSearchNs(ctxt->doc, ctxt->elem, prefix);
+		    if (ns == NULL) {
+			xsltGenericError(xsltGenericErrorContext,
+			    "xsl: pattern, no namespace bound to prefix %s\n",
+			                 prefix);
+		    } else {
+			URL = xmlStrdup(ns->href);
+		    }
+		    xmlFree(prefix);
+		}
+		xmlFree(name);
+		name = ncname;
+	    }
+	    PUSH(XSLT_OP_CHILD, name, URL);
 	} else if (xmlStrEqual(token, (const xmlChar *) "attribute")) {
-	    /* TODO: handle namespace */
 	    name = xsltScanName(ctxt);
 	    if (name == NULL) {
 		xsltGenericError(xsltGenericErrorContext,
@@ -955,7 +1100,25 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 		ctxt->error = 1;
 		goto error;
 	    }
-	    PUSH(XSLT_OP_ATTR, name, NULL);
+	    ncname = xmlSplitQName2(name, &prefix);
+	    if (ncname != NULL) {
+		if (prefix != NULL) {
+		    xmlNsPtr ns;
+
+		    ns = xmlSearchNs(ctxt->doc, ctxt->elem, prefix);
+		    if (ns == NULL) {
+			xsltGenericError(xsltGenericErrorContext,
+			    "xsl: pattern, no namespace bound to prefix %s\n",
+			                 prefix);
+		    } else {
+			URL = xmlStrdup(ns->href);
+		    }
+		    xmlFree(prefix);
+		}
+		xmlFree(name);
+		name = ncname;
+	    }
+	    PUSH(XSLT_OP_ATTR, name, URL);
 	} else {
 	    xsltGenericError(xsltGenericErrorContext,
 		"xsltCompileStepPattern : 'child' or 'attribute' expected\n");
@@ -967,8 +1130,25 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 	NEXT;
 	PUSH(XSLT_OP_ALL, token, NULL);
     } else {
-	/* TODO: handle namespace */
-	PUSH(XSLT_OP_ELEM, token, NULL);
+	ncname = xmlSplitQName2(token, &prefix);
+	if (ncname != NULL) {
+	    if (prefix != NULL) {
+		xmlNsPtr ns;
+
+		ns = xmlSearchNs(ctxt->doc, ctxt->elem, prefix);
+		if (ns == NULL) {
+		    xsltGenericError(xsltGenericErrorContext,
+			"xsl: pattern, no namespace bound to prefix %s\n",
+				     prefix);
+		} else {
+		    URL = xmlStrdup(ns->href);
+		}
+		xmlFree(prefix);
+	    }
+	    xmlFree(token);
+	    token = ncname;
+	}
+	PUSH(XSLT_OP_ELEM, token, URL);
     }
 parse_predicate:
     SKIP_BLANKS;
@@ -1116,6 +1296,8 @@ error:
 /**
  * xsltCompilePattern:
  * @pattern an XSLT pattern
+ * @doc:  the containing document
+ * @node:  the containing element
  *
  * Compile the XSLT pattern and generates a list of precompiled form suitable
  * for fast matching.
@@ -1126,7 +1308,7 @@ error:
  */
 
 xsltCompMatchPtr
-xsltCompilePattern(const xmlChar *pattern) {
+xsltCompilePattern(const xmlChar *pattern, xmlDocPtr doc, xmlNodePtr node) {
     xsltParserContextPtr ctxt = NULL;
     xsltCompMatchPtr element, first = NULL, previous = NULL;
     int current, start, end;
@@ -1137,7 +1319,7 @@ xsltCompilePattern(const xmlChar *pattern) {
 	return(NULL);
     }
 
-#ifdef DEBUG_PARSING
+#ifdef DEBUG_PATTERN
     xsltGenericDebug(xsltGenericDebugContext,
 		     "xsltCompilePattern : parsing '%s'\n", pattern);
 #endif
@@ -1145,6 +1327,8 @@ xsltCompilePattern(const xmlChar *pattern) {
     ctxt = xsltNewParserContext();
     if (ctxt == NULL)
 	return(NULL);
+    ctxt->doc = doc;
+    ctxt->elem = node;
     current = end = 0;
     while (pattern[current] != 0) {
 	start = current;
@@ -1190,6 +1374,9 @@ xsltCompilePattern(const xmlChar *pattern) {
 	    (element->steps[0].value != NULL) &&
 	    (element->steps[1].op == XSLT_OP_END)) {
 	    element->priority = 0;
+	} else if ((element->steps[0].op == XSLT_OP_ROOT) &&
+		   (element->steps[1].op == XSLT_OP_END)) {
+	    element->priority = 0;
 	} else if ((element->steps[0].op == XSLT_OP_PI) &&
 		   (element->steps[0].value != NULL) &&
 		   (element->steps[1].op == XSLT_OP_END)) {
@@ -1255,7 +1442,7 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
     if ((style == NULL) || (cur == NULL) || (cur->match == NULL))
 	return(-1);
 
-    pat = xsltCompilePattern(cur->match);
+    pat = xsltCompilePattern(cur->match, style->doc, cur->elem);
     while (pat) {
 	next = pat->next;
 	pat->next = NULL;
@@ -1265,7 +1452,7 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	    pat->mode = xmlStrdup(mode);
 	if (modeURI != NULL)
 	    pat->modeURI = xmlStrdup(modeURI);
-	if (pat->priority != XSLT_PAT_NO_PRIORITY)
+	if (cur->priority == XSLT_PAT_NO_PRIORITY)
 	    cur->priority = pat->priority;
 	else
 	    pat->priority = cur->priority;
@@ -1335,25 +1522,13 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 		    xsltFreeCompMatch(pat);
 		    return(-1);
 		}
-#ifdef DEBUG_PARSING
-		xsltGenericDebug(xsltGenericDebugContext,
-				 "xsltAddTemplate: created template hash\n");
-#endif
 		xmlHashAddEntry3(style->templatesHash, name, mode, modeURI, pat);
-#ifdef DEBUG_PARSING
-		xsltGenericDebug(xsltGenericDebugContext,
-				 "xsltAddTemplate: added new hash %s\n", name);
-#endif
 	    } else {
 		list = (xsltCompMatchPtr) xmlHashLookup3(style->templatesHash,
 							 name, mode, modeURI);
 		if (list == NULL) {
 		    xmlHashAddEntry3(style->templatesHash, name,
 				     mode, modeURI, pat);
-#ifdef DEBUG_PARSING
-		    xsltGenericDebug(xsltGenericDebugContext,
-				     "xsltAddTemplate: added new hash %s\n", name);
-#endif
 		} else {
 		    /*
 		     * Note '<=' since one must choose among the matching
@@ -1364,10 +1539,6 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 			pat->next = list;
 			xmlHashUpdateEntry3(style->templatesHash, name,
 					    mode, modeURI, pat, NULL);
-#ifdef DEBUG_PARSING
-			xsltGenericDebug(xsltGenericDebugContext,
-					 "xsltAddTemplate: added head hash for %s\n", name);
-#endif
 		    } else {
 			while (list->next != NULL) {
 			    if (list->next->priority <= pat->priority)
@@ -1402,6 +1573,12 @@ xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	    xsltFreeCompMatch(pat);
 	    return(-1);
 	}
+#ifdef DEBUG_PATTERN
+	xsltGenericDebug(xsltGenericDebugContext,
+		     "added pattern : '%s' priority %f\n",
+		         pat->template->match, pat->priority);
+#endif
+
 	pat = next;
     }
     return(0);
@@ -1549,7 +1726,6 @@ xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 		list = list->next;
 	    }
 	}
-
 	if (ret != NULL)
 	    return(ret);
 
@@ -1591,6 +1767,7 @@ xsltFreeTemplateHashes(xsltStylesheetPtr style) {
         xsltFreeCompMatchList(style->commentMatch);
 }
 
+#if 0
 /**
  * xsltMatchPattern
  * @node: a node in the source tree
@@ -1618,4 +1795,4 @@ xsltMatchPattern(xsltTransformContextPtr context,
     }
     return match;
 }
-
+#endif
