@@ -30,7 +30,6 @@
 
 #define DEBUG_TEMPLATES
 
-
 /************************************************************************
  *									*
  *			Module interfaces				*
@@ -164,7 +163,7 @@ xsltEvalTemplateString(xsltTransformContextPtr ctxt, xmlNodePtr node,
     oldInsert = ctxt->insert;
     ctxt->insert = insert;
 
-    xsltApplyOneTemplate(ctxt, node, NULL, parent->children);
+    xsltApplyOneTemplate(ctxt, node, parent->children, 0);
 
     ctxt->insert = oldInsert;
 
@@ -321,7 +320,8 @@ xsltAttrTemplateProcess(xsltTransformContextPtr ctxt, xmlNodePtr target,
             out = xsltAttrTemplateValueProcess(ctxt, in);
 	    child = xmlNewDocText(ctxt->output, out);
 	    xmlAddChild((xmlNodePtr) ret, child);
-	    xmlFree(out);
+	    if (out != NULL)
+		xmlFree(out);
 	    xmlFree(in);
 	} else
 	    ret->children = NULL;
