@@ -41,6 +41,8 @@
 
 #define DEBUG_PROCESS
 
+int xsltMaxDepth = 250;
+
 /*
  * Useful macros
  */
@@ -90,6 +92,7 @@ scope type name##Pop(xsltTransformContextPtr ctxt) {			\
  * Those macros actually generate the functions
  */
 PUSH_AND_POP(extern, xsltTemplatePtr, templ)
+PUSH_AND_POP(extern, xsltStackElemPtr, vars)
 
 /************************************************************************
  *									*
@@ -118,6 +121,10 @@ xsltNewTransformContext(xsltStylesheetPtr style, xmlDocPtr doc) {
 	return(NULL);
     }
     memset(cur, 0, sizeof(xsltTransformContext));
+
+    /*
+     * initialize the template stack
+     */
     cur->templTab = (xsltTemplatePtr *)
 	        xmlMalloc(10 * sizeof(xsltTemplatePtr));
     if (cur->templTab == NULL) {
@@ -129,6 +136,23 @@ xsltNewTransformContext(xsltStylesheetPtr style, xmlDocPtr doc) {
     cur->templNr = 0;
     cur->templMax = 5;
     cur->templ = NULL;
+
+    /*
+     * initialize the variables stack
+     */
+    cur->varsTab = (xsltStackElemPtr *)
+	        xmlMalloc(10 * sizeof(xsltStackElemPtr));
+    if (cur->varsTab == NULL) {
+        xmlGenericError(xmlGenericErrorContext,
+		"xsltNewTransformContext: out of memory\n");
+	xmlFree(cur->templTab);
+	xmlFree(cur);
+	return(NULL);
+    }
+    cur->varsNr = 0;
+    cur->varsMax = 5;
+    cur->vars = NULL;
+
     cur->style = style;
     xmlXPathInit();
     cur->xpathCtxt = xmlXPathNewContext(doc);
@@ -136,15 +160,18 @@ xsltNewTransformContext(xsltStylesheetPtr style, xmlDocPtr doc) {
         xsltGenericError(xsltGenericErrorContext,
 		"xsltNewTransformContext : xmlXPathNewContext failed\n");
 	xmlFree(cur->templTab);
+	xmlFree(cur->varsTab);
 	xmlFree(cur);
 	return(NULL);
     }
     XSLT_REGISTER_VARIABLE_LOOKUP(cur);
+    cur->xpathCtxt->nsHash = style->nsHash;
     docu = xsltNewDocument(cur, doc);
     if (docu == NULL) {
         xsltGenericError(xsltGenericErrorContext,
 		"xsltNewTransformContext : xsltNewDocument failed\n");
 	xmlFree(cur->templTab);
+	xmlFree(cur->varsTab);
 	xmlFree(cur);
 	return(NULL);
     }
@@ -163,11 +190,14 @@ void
 xsltFreeTransformContext(xsltTransformContextPtr ctxt) {
     if (ctxt == NULL)
 	return;
-    if (ctxt->xpathCtxt != NULL)
+    if (ctxt->xpathCtxt != NULL) {
+	ctxt->xpathCtxt->nsHash = NULL;
 	xmlXPathFreeContext(ctxt->xpathCtxt);
+    }
     if (ctxt->templTab != NULL)
 	xmlFree(ctxt->templTab);
-    xsltFreeVariableHashes(ctxt);
+    if (ctxt->varsTab != NULL)
+	xmlFree(ctxt->varsTab);
     xsltFreeDocuments(ctxt);
     memset(ctxt, -1, sizeof(xsltTransformContext));
     xmlFree(ctxt);
@@ -233,12 +263,19 @@ xsltCopyNode(xsltTransformContextPtr ctxt, xmlNodePtr node,
 xmlNodePtr
 xsltCopyTreeList(xsltTransformContextPtr ctxt, xmlNodePtr list,
 	     xmlNodePtr insert) {
-    xmlNodePtr copy, ret = NULL;
+    xmlNodePtr copy, ret = NULL, last;
 
     while (list != NULL) {
 	copy = xsltCopyTree(ctxt, list, insert);
-	if (ret != NULL)
-	    ret = copy;
+	if (copy != NULL) {
+	    if (ret == NULL) {
+		ret = copy;
+		last = ret;
+	    } else {
+		last->next = copy;
+		last = copy;
+	    }
+	}
 	list = list->next;
     }
     return(ret);
@@ -273,7 +310,7 @@ xsltCopyTree(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	    copy->ns = xsltGetNamespace(ctxt, node, node->ns, insert);
 	}
 	if (node->children != NULL)
-	    copy->children = xsltCopyTreeList(ctxt, node->children, copy);
+	    xsltCopyTreeList(ctxt, node->children, copy);
     } else {
 	xsltGenericError(xsltGenericErrorContext,
 		"xsltCopyTree: copy %s failed\n", node->name);
@@ -506,7 +543,9 @@ xsltCopy(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	case XML_DOCUMENT_NODE:
 	case XML_HTML_DOCUMENT_NODE:
 	case XML_ELEMENT_NODE:
-	    xsltApplyOneTemplate(ctxt, ctxt->node, NULL, inst->children);
+	    varsPush(ctxt, NULL);
+	    xsltApplyOneTemplate(ctxt, ctxt->node, inst->children, 0);
+	    xsltFreeStackElemList(varsPop(ctxt));
 	    break;
 	default:
 	    break;
@@ -514,6 +553,52 @@ xsltCopy(xsltTransformContextPtr ctxt, xmlNodePtr node,
     ctxt->insert = oldInsert;
 }
 
+/**
+ * xsltText:
+ * @ctxt:  a XSLT process context
+ * @node:  the node in the source tree.
+ * @inst:  the xslt text node
+ *
+ * Process the xslt text node on the source node
+ */
+void
+xsltText(xsltTransformContextPtr ctxt, xmlNodePtr node,
+	    xmlNodePtr inst) {
+    xmlNodePtr copy;
+
+    if (inst->children != NULL) {
+	if ((inst->children->type != XML_TEXT_NODE) ||
+	    (inst->children->next != NULL)) {
+	    xsltGenericError(xsltGenericErrorContext,
+		 "xslt:text has content problem !\n");
+	} else {
+	    xmlChar *prop;
+	    xmlNodePtr text = inst->children;
+	    
+	    copy = xmlNewDocText(ctxt->output, text->content);
+	    prop = xmlGetNsProp(inst,
+		    (const xmlChar *)"disable-output-escaping",
+				XSLT_NAMESPACE);
+	    if (prop != NULL) {
+#ifdef DEBUG_PARSING
+		xsltGenericDebug(xsltGenericDebugContext,
+		     "Disable escaping: %s\n", text->content);
+#endif
+		if (xmlStrEqual(prop, (const xmlChar *)"yes")) {
+		    copy->name = xmlStringTextNoenc;
+		} else if (!xmlStrEqual(prop,
+					(const xmlChar *)"no")){
+		    xsltGenericError(xsltGenericErrorContext,
+     "xslt:text: disable-output-escaping allow only yes or no\n");
+
+		}
+		xmlFree(prop);
+	    }
+	    xmlAddChild(ctxt->insert, copy);
+	}
+    }
+}
+
 /**
  * xsltElement:
  * @ctxt:  a XSLT process context
@@ -531,15 +616,16 @@ xsltElement(xsltTransformContextPtr ctxt, xmlNodePtr node,
     xmlChar *value = NULL;
     xmlNsPtr ns = NULL;
     xmlNodePtr copy;
+    xmlNodePtr oldInsert;
 
 
     if (ctxt->insert == NULL)
 	return;
-    if (ctxt->insert->children != NULL) {
-	xsltGenericError(xsltGenericErrorContext,
-	     "xslt:element : node has already children\n");
-	return;
-    }
+    /*
+     * stack and saves
+     */
+    oldInsert = ctxt->insert;
+
     prop = xsltEvalAttrValueTemplate(ctxt, inst, (const xmlChar *)"name");
     if (prop == NULL) {
 	xsltGenericError(xsltGenericErrorContext,
@@ -582,6 +668,7 @@ xsltElement(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	goto error;
     }
     xmlAddChild(ctxt->insert, copy);
+    ctxt->insert = copy;
     attributes = xsltEvalAttrValueTemplate(ctxt, inst,
 	                               (const xmlChar *)"use-attribute-sets");
     if (attributes != NULL) {
@@ -589,7 +676,11 @@ xsltElement(xsltTransformContextPtr ctxt, xmlNodePtr node,
         xmlFree(attributes);
     }
     
-    xsltApplyOneTemplate(ctxt, ctxt->node, NULL, inst->children);
+    varsPush(ctxt, NULL);
+    xsltApplyOneTemplate(ctxt, ctxt->node, inst->children, 0);
+    xsltFreeStackElemList(varsPop(ctxt));
+
+    ctxt->insert = oldInsert;
 
 error:
     if (prop != NULL)
@@ -896,6 +987,8 @@ xsltNumber(xsltTransformContextPtr ctxt,
 
     memset(&numdata, 0, sizeof(numdata));
     
+    numdata.doc = cur->doc;
+    numdata.node = cur;
     numdata.value = xmlGetNsProp(cur, (const xmlChar *)"value", XSLT_NAMESPACE);
     
     prop = xmlGetNsProp(cur, (const xmlChar *)"format", XSLT_NAMESPACE);
@@ -1027,7 +1120,11 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 #endif
 		oldNode = ctxt->node;
 		ctxt->node = node;
-		xsltApplyOneTemplate(ctxt, node, template, NULL);
+		templPush(ctxt, template);
+		varsPush(ctxt, NULL);
+		xsltApplyOneTemplate(ctxt, node, template->content, 1);
+		xsltFreeStackElemList(varsPop(ctxt));
+		templPop(ctxt);
 		ctxt->node = oldNode;
 	    } else /* if (ctxt->mode == NULL) */ {
 #ifdef DEBUG_PROCESS
@@ -1056,7 +1153,11 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 #endif
 		oldNode = ctxt->node;
 		ctxt->node = node;
-		xsltApplyOneTemplate(ctxt, node, template, NULL);
+		templPush(ctxt, template);
+		varsPush(ctxt, NULL);
+		xsltApplyOneTemplate(ctxt, node, template->content, 1);
+		xsltFreeStackElemList(varsPop(ctxt));
+		templPop(ctxt);
 		ctxt->node = oldNode;
 	    } else /* if (ctxt->mode == NULL) */ {
 #ifdef DEBUG_PROCESS
@@ -1086,7 +1187,11 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 
 		    oldNode = ctxt->node;
 		    ctxt->node = node;
-		    xsltApplyOneTemplate(ctxt, node, template, NULL);
+		    templPush(ctxt, template);
+		    varsPush(ctxt, NULL);
+		    xsltApplyOneTemplate(ctxt, node, template->content, 1);
+		    xsltFreeStackElemList(varsPop(ctxt));
+		    templPop(ctxt);
 		    ctxt->node = oldNode;
 		} else if (ctxt->mode == NULL) {
 		    if (attr->ns != NULL) {
@@ -1173,7 +1278,11 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 
 	    oldNode = ctxt->node;
 	    ctxt->node = node;
-	    xsltApplyOneTemplate(ctxt, node, template, NULL);
+	    templPush(ctxt, template);
+	    varsPush(ctxt, NULL);
+	    xsltApplyOneTemplate(ctxt, node, template->content, 1);
+	    xsltFreeStackElemList(varsPop(ctxt));
+	    templPop(ctxt);
 	    ctxt->node = oldNode;
 	}
 	attrs = attrs->next;
@@ -1203,7 +1312,11 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 #endif
 		    oldNode = ctxt->node;
 		    ctxt->node = node;
-		    xsltApplyOneTemplate(ctxt, node, template, NULL);
+		    templPush(ctxt, template);
+		    varsPush(ctxt, NULL);
+		    xsltApplyOneTemplate(ctxt, node, template->content, 1);
+		    xsltFreeStackElemList(varsPop(ctxt));
+		    templPop(ctxt);
 		    ctxt->node = oldNode;
 		} else /* if (ctxt->mode == NULL) */ {
 #ifdef DEBUG_PROCESS
@@ -1234,7 +1347,11 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 		    ctxt->node = cur;
 		    ctxt->xpathCtxt->contextSize = nbchild;
 		    ctxt->xpathCtxt->proximityPosition = childno;
-		    xsltApplyOneTemplate(ctxt, cur, template, NULL);
+		    templPush(ctxt, template);
+		    varsPush(ctxt, NULL);
+		    xsltApplyOneTemplate(ctxt, cur, template->content, 1);
+		    xsltFreeStackElemList(varsPop(ctxt));
+		    templPop(ctxt);
 		    ctxt->node = oldNode;
 		} else /* if (ctxt->mode == NULL) */ {
 #ifdef DEBUG_PROCESS
@@ -1265,7 +1382,11 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 		    ctxt->node = cur;
 		    ctxt->xpathCtxt->contextSize = nbchild;
 		    ctxt->xpathCtxt->proximityPosition = childno;
-		    xsltApplyOneTemplate(ctxt, cur, template, NULL);
+		    templPush(ctxt, template);
+		    varsPush(ctxt, NULL);
+		    xsltApplyOneTemplate(ctxt, cur, template->content, 1);
+		    xsltFreeStackElemList(varsPop(ctxt));
+		    templPop(ctxt);
 		    ctxt->node = oldNode;
 		}
 		break;
@@ -1298,7 +1419,11 @@ xsltApplyImports(xsltTransformContextPtr ctxt, xmlNodePtr node,
     }
     template = xsltGetTemplate(ctxt, node, ctxt->templ->style);
     if (template != NULL) {
-	xsltApplyOneTemplate(ctxt, node, template, NULL);
+	templPush(ctxt, template);
+	varsPush(ctxt, NULL);
+	xsltApplyOneTemplate(ctxt, node, template->content, 1);
+	xsltFreeStackElemList(varsPop(ctxt));
+	templPop(ctxt);
     }
 }
 
@@ -1319,7 +1444,7 @@ xsltCallTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
     xmlNsPtr ns = NULL;
     xsltTemplatePtr template;
     xmlNodePtr cur = NULL;
-    int has_param = 0;
+    xsltStackElemPtr params = NULL, param;
 
 
     if (ctxt->insert == NULL)
@@ -1350,19 +1475,21 @@ xsltCallTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	template = xsltFindTemplate(ctxt, ncname, NULL);
     if (template == NULL) {
 	xsltGenericError(xsltGenericDebugContext,
-	     "xslt:call-template: template %s not found\n", cur->name);
+	     "xslt:call-template: template %s not found\n", ncname);
 	goto error;
     }
+    /*
+     * Create a new frame but block access to variables
+     */
+    templPush(ctxt, template);
     cur = inst->children;
     while (cur != NULL) {
-	CHECK_STOPPEDE;
+	if (ctxt->state == XSLT_STATE_STOPPED) break;
 	if (IS_XSLT_ELEM(cur)) {
 	    if (IS_XSLT_NAME(cur, "with-param")) {
-		if (has_param == 0) {
-		    xsltPushStack(ctxt);
-		    has_param = 1;
-		}
-		xsltParseStylesheetParam(ctxt, cur);
+		param = xsltParseStylesheetCallerParam(ctxt, cur);
+		param->next = params;
+		params = param;
 	    } else {
 		xsltGenericError(xsltGenericDebugContext,
 		     "xslt:call-template: misplaced xslt:%s\n", cur->name);
@@ -1373,11 +1500,12 @@ xsltCallTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	}
 	cur = cur->next;
     }
-    xsltApplyOneTemplate(ctxt, node, template, NULL);
+    varsPush(ctxt, params);
+    xsltApplyOneTemplate(ctxt, node, template->content, 1);
+    xsltFreeStackElemList(varsPop(ctxt));
+    templPop(ctxt);
 
 error:
-    if (has_param == 1)
-	xsltPopStack(ctxt);
     if (prop != NULL)
         xmlFree(prop);
     if (ncname != NULL)
@@ -1602,35 +1730,44 @@ error:
  * xsltApplyOneTemplate:
  * @ctxt:  a XSLT process context
  * @node:  the node in the source tree.
- * @templ:  the template
  * @list:  the template replacement nodelist
+ * @real: is this a real template processing
  *
  * Process the apply-templates node on the source node
  */
 void
 xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
-	             xsltTemplatePtr templ, xmlNodePtr list) {
+	             xmlNodePtr list, int real) {
     xmlNodePtr cur = NULL, insert, copy = NULL;
     xmlNodePtr oldInsert;
+    xmlNodePtr oldCurrent;
     xmlAttrPtr attrs;
-    int has_variables = 0;
 
-    if ((templ == NULL) && (list == NULL))
+    if (list == NULL)
 	return;
     CHECK_STOPPED;
 
+    if (ctxt->templNr >= xsltMaxDepth) {
+	xsltGenericError(xsltGenericErrorContext,
+		"xsltApplyOneTemplate: loop found ???\n");
+	xsltGenericError(xsltGenericErrorContext,
+		"try increasing xsltMaxDepth (--maxdepth)\n");
+	xsltDebug(ctxt, node);
+	return;
+    }
+
     /*
      * stack and saves
      */
-    if (templ != NULL)
-	templPush(ctxt, templ);
     oldInsert = insert = ctxt->insert;
+    if (real) {
+	oldCurrent = ctxt->current;
+	ctxt->current = node;
+    }
 
     /*
      * Insert all non-XSLT nodes found in the template
      */
-    if ((list == NULL) && (templ != NULL))
-	list = templ->content;
     cur = list;
     while (cur != NULL) {
 	/*
@@ -1641,11 +1778,8 @@ xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	    xsltGenericDebug(xsltGenericDebugContext,
 		 "xsltApplyOneTemplate: insert == NULL !\n");
 #endif
-	    if (has_variables != 0) {
-		xsltPopStack(ctxt);
-	    }
-	    if (templ != NULL)
-		templPop(ctxt);
+	    if (real)
+		ctxt->current = oldCurrent;
 	    return;
 	}
 
@@ -1690,6 +1824,12 @@ xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 		ctxt->insert = insert;
 		xsltElement(ctxt, node, cur);
 		ctxt->insert = oldInsert;
+	    } else if (IS_XSLT_NAME(cur, "text")) {
+#ifdef DEBUG_PROCESS
+		xsltGenericDebug(xsltGenericDebugContext,
+		     "xsltApplyOneTemplate: found xslt:text, not expected\n");
+#endif
+		xsltText(ctxt, node, cur);
 	    } else if (IS_XSLT_NAME(cur, "comment")) {
 		ctxt->insert = insert;
 		xsltComment(ctxt, node, cur);
@@ -1703,32 +1843,18 @@ xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 		xsltProcessingInstruction(ctxt, node, cur);
 		ctxt->insert = oldInsert;
 	    } else if (IS_XSLT_NAME(cur, "variable")) {
-		if (has_variables == 0) {
-		    xsltPushStack(ctxt);
-		    has_variables = 1;
-		}
 		xsltParseStylesheetVariable(ctxt, cur);
 	    } else if (IS_XSLT_NAME(cur, "param")) {
-		if (has_variables == 0) {
-		    xsltPushStack(ctxt);
-		    has_variables = 1;
-		}
 		xsltParseStylesheetParam(ctxt, cur);
 	    } else if (IS_XSLT_NAME(cur, "call-template")) {
 		ctxt->insert = insert;
-		if (has_variables == 0) {
-		    xsltPushStack(ctxt);
-		    has_variables = 1;
-		}
 		xsltCallTemplate(ctxt, node, cur);
 		ctxt->insert = oldInsert;
 	    } else if (IS_XSLT_NAME(cur, "message")) {
 		xsltMessage(ctxt, node, cur);
 	    } else {
-#ifdef DEBUG_PROCESS
 		xsltGenericError(xsltGenericDebugContext,
 		     "xsltApplyOneTemplate: found xslt:%s\n", cur->name);
-#endif
 		TODO
 	    }
 	    CHECK_STOPPED;
@@ -1812,11 +1938,8 @@ skip_children:
 	    }
 	} while (cur != NULL);
     }
-    if (has_variables != 0) {
-	xsltPopStack(ctxt);
-    }
-    if (templ != NULL)
-	templPop(ctxt);
+    if (real)
+	ctxt->current = oldCurrent;
 }
 
 /**
@@ -1899,7 +2022,9 @@ xsltChoose(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	    "xsl:when: test evaluate to %d\n", doit);
 #endif
 	if (doit) {
-	    xsltApplyOneTemplate(ctxt, ctxt->node, NULL, when->children);
+	    varsPush(ctxt, NULL);
+	    xsltApplyOneTemplate(ctxt, ctxt->node, when->children, 0);
+	    xsltFreeStackElemList(varsPop(ctxt));
 	    goto done;
 	}
 	if (xpathParserCtxt != NULL)
@@ -1914,7 +2039,9 @@ xsltChoose(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	replacement = replacement->next;
     }
     if (IS_XSLT_ELEM(replacement) && (IS_XSLT_NAME(replacement, "otherwise"))) {
-	xsltApplyOneTemplate(ctxt, ctxt->node, NULL, replacement->children);
+	varsPush(ctxt, NULL);
+	xsltApplyOneTemplate(ctxt, ctxt->node, replacement->children, 0);
+	xsltFreeStackElemList(varsPop(ctxt));
 	replacement = replacement->next;
     }
     if (replacement != NULL) {
@@ -1998,7 +2125,9 @@ xsltIf(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	"xsltIf: test evaluate to %d\n", doit);
 #endif
     if (doit) {
-	xsltApplyOneTemplate(ctxt, node, NULL, inst->children);
+	varsPush(ctxt, NULL);
+	xsltApplyOneTemplate(ctxt, node, inst->children, 0);
+	xsltFreeStackElemList(varsPop(ctxt));
     }
 
 error:
@@ -2090,7 +2219,9 @@ xsltForEach(xsltTransformContextPtr ctxt, xmlNodePtr node,
     for (i = 0;i < list->nodeNr;i++) {
 	ctxt->node = list->nodeTab[i];
 	ctxt->xpathCtxt->proximityPosition = i + 1;
-	xsltApplyOneTemplate(ctxt, list->nodeTab[i], NULL, replacement);
+	varsPush(ctxt, NULL);
+	xsltApplyOneTemplate(ctxt, list->nodeTab[i], replacement, 0);
+	xsltFreeStackElemList(varsPop(ctxt));
     }
     ctxt->nodeList = oldlist;
     ctxt->xpathCtxt->contextSize = oldContextSize;
@@ -2146,7 +2277,11 @@ xsltProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	     "xsltProcessOneNode: applying template for attribute %s\n",
 	                 node->name);
 #endif
-	xsltApplyOneTemplate(ctxt, node, template, NULL);
+	templPush(ctxt, template);
+	varsPush(ctxt, NULL);
+	xsltApplyOneTemplate(ctxt, node, template->content, 1);
+	xsltFreeStackElemList(varsPop(ctxt));
+	templPop(ctxt);
     } else {
 #ifdef DEBUG_PROCESS
 	if (node->type == XML_DOCUMENT_NODE)
@@ -2158,7 +2293,11 @@ xsltProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 #endif
 	oldNode = ctxt->node;
 	ctxt->node = node;
-	xsltApplyOneTemplate(ctxt, node, template, NULL);
+	templPush(ctxt, template);
+	varsPush(ctxt, NULL);
+	xsltApplyOneTemplate(ctxt, node, template->content, 1);
+	xsltFreeStackElemList(varsPop(ctxt));
+	templPop(ctxt);
 	ctxt->node = oldNode;
     }
 }
@@ -2184,7 +2323,6 @@ xsltApplyStylesheet(xsltStylesheetPtr style, xmlDocPtr doc) {
     ctxt = xsltNewTransformContext(style, doc);
     if (ctxt == NULL)
 	return(NULL);
-    xsltEvalGlobalVariables(ctxt);
     if ((style->method != NULL) &&
 	(!xmlStrEqual(style->method, (const xmlChar *) "xml"))) {
 	if (xmlStrEqual(style->method, (const xmlChar *) "html")) {
@@ -2226,8 +2364,11 @@ xsltApplyStylesheet(xsltStylesheetPtr style, xmlDocPtr doc) {
      */
     ctxt->output = res;
     ctxt->insert = (xmlNodePtr) res;
+    xsltEvalGlobalVariables(ctxt);
     ctxt->node = (xmlNodePtr) doc;
+    varsPush(ctxt, NULL);
     xsltProcessOneNode(ctxt, ctxt->node);
+    xsltFreeStackElemList(varsPop(ctxt));
 
 
     if ((ctxt->type = XSLT_OUTPUT_XML) &&
