@@ -19,7 +19,6 @@
 extern "C" {
 #endif
 
-
 /*
  * The in-memory structure corresponding to an XSLT Variable
  * or Param
@@ -54,12 +53,13 @@ struct _xsltTemplate {
     struct _xsltTemplate *next;/* chained list sorted by priority */
     struct _xsltStylesheet *style;/* the containing stylesheet */
     xmlChar *match;	/* the matching string */
-    int priority;	/* as given from the stylesheet, not computed */
+    float priority;	/* as given from the stylesheet, not computed */
     xmlChar *name;	/* the local part of the name QName */
     xmlChar *nameURI;	/* the URI part of the name QName */
     xmlChar *mode;	/* the local part of the mode QName */
     xmlChar *modeURI;	/* the URI part of the mode QName */
     xmlNodePtr content;	/* the template replacement value */
+    xmlNodePtr elem;	/* the source element */
 };
 
 /*
@@ -150,6 +150,11 @@ struct _xsltStylesheet {
      */
     xmlHashTablePtr attributeSets;/* the attribute sets hash tables */
 
+    /*
+     * Attribute sets
+     */
+    xmlHashTablePtr nsHash;     /* the set of namespaces in use */
+
     /*
      * Key definitions
      */
@@ -200,24 +205,37 @@ struct _xsltTransformContext {
     int              templMax;		/* Size of the templtes stack */
     xsltTemplatePtr *templTab;		/* the template stack */
 
+    xsltStackElemPtr  vars;		/* the current variable list */
+    int               varsNr;		/* Nb of variable list in the stack */
+    int               varsMax;		/* Size of the variable list stack */
+    xsltStackElemPtr *varsTab;		/* the variable list stack */
+
+    /*
+     * Extensions
+     */
+    xmlHashTablePtr   extFunctions;	/* the extension functions */
+    xmlHashTablePtr   extElements;	/* the extension elements */
+
     const xmlChar *mode;		/* the current mode */
     const xmlChar *modeURI;		/* the current mode URI */
 
     xsltDocumentPtr docList;		/* the document list */
 
     xsltDocumentPtr document;		/* the current document */
-    xmlNodePtr node;			/* the current node */
+    xmlNodePtr node;			/* the node being processed */
     xmlNodeSetPtr nodeList;		/* the current node list */
+    xmlNodePtr current;			/* the current node */
 
     xmlDocPtr output;			/* the resulting document */
     xmlNodePtr insert;			/* the insertion node */
 
     xmlXPathContextPtr xpathCtxt;	/* the XPath context */
-    void *variablesHash;		/* hash table or wherever variables
-				   	   informations are stored */
     xsltTransformState state;		/* the current state */
 };
 
+typedef void (*xsltTransformFunction) (xsltTransformContextPtr ctxt,
+	                               xmlNodePtr node, xmlNodePtr inst);
+
 #define CHECK_STOPPED if (ctxt->state == XSLT_STATE_STOPPED) return;
 #define CHECK_STOPPEDE if (ctxt->state == XSLT_STATE_STOPPED) goto error;
 #define CHECK_STOPPED0 if (ctxt->state == XSLT_STATE_STOPPED) return(0);
