@@ -40,6 +40,49 @@
  */
 void 
 xsltDebug(xsltTransformContextPtr ctxt, xmlNodePtr node) {
+    int i, j;
+
+    fprintf(stderr, "Templates:\n");
+    for (i = 0, j = ctxt->templNr - 1;((i < 5) && (j >= 0));i++,j--) {
+	fprintf(stderr, "#%d ", i);
+	if (ctxt->templTab[j]->name != NULL)
+	    fprintf(stderr, "name %s ", ctxt->templTab[j]->name);
+	if (ctxt->templTab[j]->match != NULL)
+	    fprintf(stderr, "name %s ", ctxt->templTab[j]->match);
+	if (ctxt->templTab[j]->mode != NULL)
+	    fprintf(stderr, "name %s ", ctxt->templTab[j]->mode);
+	fprintf(stderr, "\n");
+    }
+    fprintf(stderr, "Variables:\n");
+    for (i = 0, j = ctxt->varsNr - 1;((i < 10) && (j >= 0));i++,j--) {
+	xsltStackElemPtr cur;
+
+	if (ctxt->varsTab[j] == NULL)
+	    continue;
+	fprintf(stderr, "#%d\n", i);
+	cur = ctxt->varsTab[j];
+	while (cur != NULL) {
+	    if (cur->type == XSLT_ELEM_VARIABLE)
+		fprintf(stderr, "var ");
+	    else if (cur->type == XSLT_ELEM_PARAM)
+		fprintf(stderr, "param ");
+	    else {
+		fprintf(stderr, "corrupted !!!\n");
+	    }
+	    if (cur->name != NULL)
+		fprintf(stderr, "%s ", cur->name);
+	    else
+		fprintf(stderr, "noname !!!!");
+	    if (cur->value != NULL) {
+		xmlXPathDebugDumpObject(stderr, cur->value, 1);
+	    } else {
+		fprintf(stderr, "NULL !!!!");
+	    }
+	    fprintf(stderr, "\n");
+	    cur = cur->next;
+	}
+	
+    }
 }
 
 /************************************************************************
@@ -78,7 +121,11 @@ xsltMessage(xsltTransformContextPtr ctxt, xmlNodePtr node, xmlNodePtr inst) {
     }
     message = xsltEvalTemplateString(ctxt, node, inst);
     if (message != NULL) {
+	int len = xmlStrlen(message);
+
 	xsltGenericError(xsltGenericErrorContext, (const char *)message);
+	if ((len > 0) && (message[len - 1] != '\n'))
+	    xsltGenericError(xsltGenericErrorContext, "\n");
 	xmlFree(message);
     }
     if (terminate)
@@ -232,7 +279,7 @@ xsltDocumentSortFunction(xmlNodeSetPtr list) {
 void	
 xsltSortFunction(xmlNodeSetPtr list, xmlXPathObjectPtr *results,
 		 int descending, int number) {
-    int i, j;
+    int i, j, incr;
     int len, tst;
     xmlNodePtr node;
     xmlXPathObjectPtr tmp;
@@ -242,30 +289,37 @@ xsltSortFunction(xmlNodeSetPtr list, xmlXPathObjectPtr *results,
     len = list->nodeNr;
     if (len <= 1)
 	return;
-    /* TODO: sort is really not optimized, does it needs to ? */
-    for (i = 0;i < len -1;i++) {
-	for (j = i + 1; j < len; j++) {
+
+    /* Shell's sort of node-set */
+    for (incr = len / 2; incr > 0; incr /= 2) {
+	for (i = incr; i < len; i++) {
+	    j = i - incr;
 	    if (results[i] == NULL)
-		tst = 0;
-	    else if (results[j] == NULL)
-		tst = 1;
-	    else if (number) {
-		tst = (results[i]->floatval > results[j]->floatval);
-		if (descending)
-		    tst = !tst;
-	    } else {
-		tst = (xmlStrcmp(results[i]->stringval,
-			         results[j]->stringval)) > 0; 
-		if (descending)
-		    tst = !tst;
-	    }
-	    if (tst) {
-		tmp = results[i];
-		results[i] = results[j];
-		results[j] = tmp;
-		node = list->nodeTab[i];
-		list->nodeTab[i] = list->nodeTab[j];
-		list->nodeTab[j] = node;
+		continue;
+	    
+	    while (j >= 0) {
+		if (results[j] == NULL)
+		    tst = 1;
+		else if (number) {
+		    tst = (results[j]->floatval > results[j + incr]->floatval);
+		    if (descending)
+			tst = !tst;
+		} else {
+		    tst = (xmlStrcmp(results[j]->stringval,
+				     results[j + incr]->stringval)) > 0; 
+		    if (descending)
+			tst = !tst;
+		}
+		if (tst) {
+		    tmp = results[j];
+		    results[j] = results[j + incr];
+		    results[j + incr] = tmp;
+		    node = list->nodeTab[j];
+		    list->nodeTab[j] = list->nodeTab[j + incr];
+		    list->nodeTab[j + incr] = node;
+		    j -= incr;
+		} else
+		    break;
 	    }
 	}
     }
